package jp.ne.motoki.android.hellolivewallpaper;

import static jp.ne.motoki.android.hellolivewallpaper.Utils.log_debug;
import static jp.ne.motoki.android.hellolivewallpaper.Utils.log_error;
import static jp.ne.motoki.android.hellolivewallpaper.Utils.log_warn;

import javax.microedition.khronos.egl.EGL10;
import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.egl.EGLContext;
import javax.microedition.khronos.egl.EGLDisplay;
import javax.microedition.khronos.egl.EGLSurface;
import javax.microedition.khronos.opengles.GL10;

import android.content.Context;
import android.os.Handler;
import android.os.Message;
import android.view.SurfaceHolder;
import android.widget.Toast;

public class GLEngineSurface extends Thread {
    
    private final Handler ERROR_HANDLER = new Handler() {
        @Override
        public void handleMessage(Message msg) {
            showToast();
        }
    };
    
    private final Context CONTEXT;

    /**
     * 描画対象のSurfaceHolder。
     */
    private final SurfaceHolder HOLDER;

    /**
     * EGLインターフェース。
     */
    private EGL10 egl = null;
    
    /**
     * ディスプレイ。
     */
    private EGLDisplay eglDisplay = null;

    /**
     * コンフィグ情報。
     */
    private EGLConfig eglConfig = null;

    /**
     * GLコンテキスト。
     */
    private EGLContext eglContext = null;
    
    /**
     * サーフェイス。
     */
    private EGLSurface eglSurface = null;

    /**
     * GL用インターフェース。
     */
    private GL10 gl10 = null;

    /**
     * 描画先サーフェイスの幅・高さ
     */
    int windowWidth = -1, windowHeight = -1;
    
    private boolean destroy = false;
    private boolean pause = false;

    public GLEngineSurface(Context context, SurfaceHolder holder) {
        if (context == null) {
            throw new NullPointerException("context == null");
        }
        if (holder == null) {
            throw new NullPointerException("holder == null");
        }
        
        CONTEXT = context;
        HOLDER = holder;
    }
    
    public void onDestroy() {
        log_debug("onDestroy");
        synchronized(this) {
            destroy = true;
            notify();
        }
        
        try {
            join();
        } catch (InterruptedException e) {
            log_warn("join interrupted");
            Thread.currentThread().interrupt();
        }
    }

    public synchronized void onPause() {
        log_debug("onPause");
        pause = true;
        notify();
    }

    public synchronized void onResume() {
        log_debug("onResume");
        pause = false;
        notify();
    }

    @Override
    public void run() {
        GLEngineRender render = new GLEngineRender(CONTEXT);
        try {
            initialize();
        } catch (HelloWallpaperException e) {
            // TODO 仮実装
            log_error("initialize", e);
            ERROR_HANDLER.obtainMessage().sendToTarget();
            dispose();
            return;
        }
        render.onSurfaceChanged(gl10, windowWidth, windowHeight);
        while (true) {
            synchronized (this) {
                if (destroy) {
                    break;
                }
                if (pause) {
                    try {
                        wait();
                    } catch (InterruptedException e) {
                        log_warn("wait interrupted");
                    }
                    continue;
                }
            }
            gl10.glViewport(0, 0, windowWidth, windowHeight);
            render.onDrawFrame(gl10);
            egl.eglSwapBuffers(eglDisplay, eglSurface);
        }
        dispose();
    }

    /**
     * OpenGL 初期化処理を行う
     * egl初期化処理を行ったスレッドのみがOpneGL ES を利用できる
     */
    private void initialize() throws HelloWallpaperException {
        egl = (EGL10) EGLContext.getEGL();
        
        eglDisplay = getDisplay();
        initializeEGL();
        eglConfig = getConfig();
        eglContext = createContext();
        eglSurface = eglCreateWindowSurface();
        gl10 = getGL();
        makeCurrent();
    }
    
    /**
     * 描画先ディスプレイ確保
     */
    private EGLDisplay getDisplay() {
        return egl.eglGetDisplay(EGL10.EGL_DEFAULT_DISPLAY);
    }
    
    /**
     * EGL 初期化
     * @return 初期化成功の場合{@code true}。
     * @throws HelloWallpaperException
     */
    private void initializeEGL() throws HelloWallpaperException {
        int[] version = { -1, -1 };
        if (!egl.eglInitialize(eglDisplay, version)) {
            throw new HelloWallpaperException("!eglInitialize");
        }
        log_debug(String.format(
                "EGL version is %d.%d", version[0], version[1]));
    }
    
    /**
     * コンフィグ取得
     * @return コンフィグ。取得に失敗した場合、{@code null}。
     * @throws HelloWallpaperException
     */
    private EGLConfig getConfig() throws HelloWallpaperException {
        EGLConfig[] configs = new EGLConfig[1];
        int[] num = new int[1];
        
        int[] spec = { EGL10.EGL_NONE };
        if (!egl.eglChooseConfig(eglDisplay, spec, configs, 1, num)) {
            throw new HelloWallpaperException("!eglChooseConfig");
        }
        return configs[0];
    }
    
    /**
     * レンダリングコンテキスト作成
     * @return
     * @throws HelloWallpaperException
     */
    private EGLContext createContext() throws HelloWallpaperException {
        EGLContext context = egl.eglCreateContext(
                eglDisplay, eglConfig, EGL10.EGL_NO_CONTEXT, null);
        if (context == EGL10.EGL_NO_CONTEXT) {
            throw new HelloWallpaperException(
                    "context == EGL10.EGL_NO_CONTEXT");
        }
        return context;
    }
    
    /**
     * 描画先サーフェイス作成
     * @return
     */
    private EGLSurface eglCreateWindowSurface() throws HelloWallpaperException {
        egl.eglMakeCurrent(
                eglDisplay, EGL10.EGL_NO_SURFACE,
                EGL10.EGL_NO_SURFACE, EGL10.EGL_NO_CONTEXT);
        EGLSurface surface = egl.eglCreateWindowSurface(
                eglDisplay, eglConfig, HOLDER, null);
        if (surface == EGL10.EGL_NO_SURFACE) {
            throw new HelloWallpaperException(
                    "surface == EGL10.EGL_NO_SURFACE");
        }
        return surface;
    }
    
    private GL10 getGL() {
        return (GL10) eglContext.getGL();
    }
    
    /**
     * サーフェイスとコンテキストを結びつける
     * @throws HelloWallpaperException
     */
    private void makeCurrent() throws HelloWallpaperException {
        if (!egl.eglMakeCurrent(
                eglDisplay, eglSurface, eglSurface, eglContext)) {
            throw new HelloWallpaperException("!eglMakeCurrent");
        }
    }

    /**
     * GLの終了を行う。
     */
    private void dispose() {
        log_debug("dispose");
        destroySurface();
        destroyContext();
        terminateDisplayConnection();
    }
    
    /**
     * サーフェイス破棄
     */
    private void destroySurface() {
        if (eglSurface != null) {
            egl.eglMakeCurrent(
                    eglDisplay, EGL10.EGL_NO_SURFACE,
                    EGL10.EGL_NO_SURFACE, EGL10.EGL_NO_CONTEXT);
            egl.eglDestroySurface(eglDisplay, eglSurface);
            eglSurface = null;
        }
    }
    
    /**
     * レンダリングコンテキスト破棄
     */
    private void destroyContext() {
        if (eglContext != null) {
            egl.eglDestroyContext(eglDisplay, eglContext);
            eglContext = null;
        }
    }
    
    /**
     * ディスプレイコネクション破棄
     */
    private void terminateDisplayConnection() {
        if (eglDisplay != null) {
            egl.eglTerminate(eglDisplay);
            eglDisplay = null;
        }
    }
    
    private void showToast() {
        Toast.makeText(
                CONTEXT, "Initialization failed", Toast.LENGTH_LONG).show();
    }
}